
#include "interpolate.h"

#include "assert.h"

void write_silence (FILE *ofile, int num_samples) {
    short  missing_pkt [num_samples];
    int i;
    for (i = 0; i < num_samples; i++) {
        missing_pkt[i] = 0;
    }
    fwrite (missing_pkt, 1, num_samples * 2, ofile);
}

void distribute_last_packet (short  value, short  *packet, int num_samples) {
    short  lastElement = value;
    int lastPosShaping = 0.30 * num_samples;
    double difference = packet[0] - lastElement;
    double step_size = difference / lastPosShaping;
    int i;
    for (i = 0; i < lastPosShaping; i++) {
        packet[i] -= step_size * (lastPosShaping - i);
    }
}

void distribute_both_packets (short  *packet1, short  *packet2, int num_samples) {
    int lastPosShaping = 0.15 * num_samples;
    double mediumPoint = (packet2[0] + packet1[num_samples - 1]) / 2;
    double difference1 = mediumPoint - packet1[num_samples - 1];
    double difference2 = packet2[0] - mediumPoint;
    double step_size1 = difference1 / lastPosShaping;
    double step_size2 = difference2 / lastPosShaping;
    int i;
    for (i = num_samples - lastPosShaping; i < num_samples; i++) {
        packet1[i] += step_size1 * (i - num_samples + lastPosShaping);
    }
    for (i = 0; i < lastPosShaping; i++) {
        packet2[i] -= step_size2 * (lastPosShaping - i);
    }
}

void recv_packet (int seqno, int len, char *data, FILE *ofile, int strategy) {
    static int prev_seqno = -1;
    static short  *prev_samples = 0;
    static long  stream_position = 0;
    short  *samples = (short  *) data;
    int num_samples = len / 2;
    printf ("recv_packet: seqno=%d\n", seqno);
    if (prev_seqno != -1 && (prev_seqno + 1 != seqno)) {
        int i, missing_seqno;
        printf ("s=%d\n", strategy);
        switch (strategy) {
        case SILENCE :
            {
                missing_seqno = prev_seqno + 1;
                while (missing_seqno < seqno) {
                    write_silence (ofile, num_samples);
                    missing_seqno++;
                }
                break;
            }
        case REPEAT_PREV :
            {
                fwrite (prev_samples, 2, num_samples, ofile);
                missing_seqno = prev_seqno + 2;
                while (missing_seqno < seqno) {
                    write_silence (ofile, num_samples);
                    missing_seqno++;
                }
                break;
            }
        case REPEAT_PREV_NEXT :
            {
                int nrOfLostPackets = seqno - prev_seqno - 1;
                switch (nrOfLostPackets) {
                case 1 :
                    {
                        fwrite (prev_samples + num_samples / 2, 2, num_samples / 2, ofile);
                        fwrite (samples, 2, num_samples / 2, ofile);
                        break;
                    }
                default :
                    {
                        fwrite (prev_samples, 2, num_samples, ofile);
                        nrOfLostPackets = nrOfLostPackets - 2;
                        while (nrOfLostPackets > 0) {
                            write_silence (ofile, num_samples);
                            nrOfLostPackets--;
                        }
                        fwrite (samples, 2, num_samples, ofile);
                        break;
                    }
                }
                break;
            }
        case INTERPOLATE :
            {
                int *dctPrev = dct (prev_samples);
                int *dctCurr = dct (samples);
                int nrOfLostPackets = seqno - prev_seqno - 1;
                int kk;
                if (nrOfLostPackets == 1) {
                    int *odata = malloc (num_samples * sizeof (int));
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.5 * dctPrev[kk] + 0.5 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata[kk] = (int) intermediate;
                    }
                    short  *newPacket = idct (odata);
                    distribute_both_packets (prev_samples, newPacket, num_samples);
                    fseek (ofile, stream_position, SEEK_SET);
                    fwrite (prev_samples, 2, num_samples, ofile);
                    distribute_both_packets (newPacket, samples, num_samples);
                    fwrite (newPacket, 2, num_samples, ofile);
                }
                else if (nrOfLostPackets == 2) {
                    int *odata1 = malloc (num_samples * sizeof (int));
                    int *odata2 = malloc (num_samples * sizeof (int));
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.7 * dctPrev[kk] + 0.3 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata1[kk] = (int) intermediate;
                    }
                    short  *newPacket1 = idct (odata1);
                    distribute_both_packets (prev_samples, newPacket1, num_samples);
                    fseek (ofile, stream_position, SEEK_SET);
                    fwrite (prev_samples, 2, num_samples, ofile);
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.3 * dctPrev[kk] + 0.7 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata2[kk] = (int) intermediate;
                    }
                    short  *newPacket2 = idct (odata2);
                    distribute_both_packets (newPacket1, newPacket2, num_samples);
                    distribute_both_packets (newPacket2, samples, num_samples);
                    fwrite (newPacket1, 2, num_samples, ofile);
                    fwrite (newPacket2, 2, num_samples, ofile);
                }
                else if (nrOfLostPackets >= 3) {
                    int *odata1 = malloc (num_samples * sizeof (int));
                    int *odata2 = malloc (num_samples * sizeof (int));
                    int *odata3 = malloc (num_samples * sizeof (int));
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.9 * dctPrev[kk] + 0.1 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata1[kk] = (int) intermediate;
                    }
                    short  *newPacket1 = idct (odata1);
                    distribute_both_packets (prev_samples, newPacket1, num_samples);
                    fseek (ofile, stream_position, SEEK_SET);
                    fwrite (prev_samples, 2, num_samples, ofile);
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.5 * dctPrev[kk] + 0.5 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata2[kk] = (int) intermediate;
                    }
                    short  *newPacket2 = idct (odata2);
                    distribute_both_packets (newPacket1, newPacket2, num_samples);
                    for (kk = 0; kk < num_samples; kk++) {
                        double intermediate = abs (0.1 * dctPrev[kk] + 0.9 * dctCurr[kk]);
                        int sign = (rand ()) % 2;
                        if (sign = 1) {
                            intermediate = 0 - intermediate;
                        }
                        odata3[kk] = (int) intermediate;
                    }
                    short  *newPacket3 = idct (odata3);
                    distribute_both_packets (newPacket2, newPacket3, num_samples);
                    fwrite (newPacket1, 2, num_samples, ofile);
                    fwrite (newPacket2, 2, num_samples, ofile);
                    nrOfLostPackets = nrOfLostPackets - 3;
                    if (nrOfLostPackets <= 0) {
                        distribute_both_packets (newPacket3, samples, num_samples);
                        fwrite (newPacket3, 2, num_samples, ofile);
                    }
                    else {
                        fwrite (newPacket3, 2, num_samples, ofile);
                        while (nrOfLostPackets > 0) {
                            write_silence (ofile, num_samples);
                            nrOfLostPackets--;
                        }
                    }
                }
                break;
            }
        default :
            abort ();
        }
    }
    stream_position = ftell (ofile);
    fwrite (samples, 1, num_samples * 2, ofile);
    if (prev_samples != 0)
        free (prev_samples);
    prev_samples = samples;
    prev_seqno = seqno;
};

